"""
Wiggelen command line interface.

.. moduleauthor:: Martijn Vermaat <martijn@vermaat.name>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import division

import argparse
import importlib
import re
import sys

from .wiggle import fill, walk, write
from .index import index
from .merge import merge, mergers
from .distance import metrics, distance
from .transform import (backward_divided_difference,
                        forward_divided_difference,
                        central_divided_difference)
from . import intervals

# Python 3 compatibility.
try:
    from itertools import imap, ifilter
    map_ = imap
    filter_ = ifilter
except ImportError:
    map_ = map
    filter_ = filter

# Plotting only if Matplotlib is installed.
try:
    from matplotlib import pyplot
    from .plot import plot
except ImportError:
    matplotlib = plot = None


def log(message):
    sys.stderr.write(message + '\n')


def abort(message=None):
    if message:
        log('error: ' + message)
    sys.exit(1)


def read_regions(regions):
    """
    Ad-hoc parsing of regions from a BED track.
    """
    result = {}
    for line in regions:
        if line.startswith('track'):
            continue
        region, start, stop = line.strip().split('\t')[:3]
        result[region] = int(start) + 1, int(stop)
    return result


def index_track(track):
    """
    Build index for wiggle track.
    """
    # Todo: This will not rebuild the index if it already exists.
    idx, filename = index(track, force=True)
    if filename is None:
        abort('Could not write index file')


def sort_track(track, name=None, description=None):
    """
    Sort wiggle track regions alphabetically.
    """
    if name is None and hasattr(track, 'name'):
        name = 'Sorted %s' % track.name

    write(walk(track, force_index=True), name=name, description=description)


def scale_track(track, factor=0.1, name=None, description=None):
    """
    Scale values in a wiggle track.
    """
    if name is None and hasattr(track, 'name'):
        name = 'Scaled %s' % track.name

    scale = lambda (r, p, v): (r, p, v * factor)
    write(map_(scale, walk(track)), name=name, description=description)


def fill_track(track, genome=None, filler='0', only_edges=False,
               only_genome=False, name=None, description=None):
    """
    Fill in undefined positions in a wiggle track.
    """
    if name is None and hasattr(track, 'name'):
        name = 'Filled %s' % track.name

    walker = walk(track)

    if genome is not None:
        genome = read_regions(genome)

        if only_genome:
            # Todo: This is a bit of a hack to keep only positions defined by
            #   the genome. We might want to generalize this for use in other
            #   functions and to multiple definitions per region (currently,
            #   the genome definition can have only one definition per
            #   region).
            def in_genome((region, position, value)):
                try:
                    start, stop = genome[region]
                    return start <= position <= stop
                except KeyError:
                    return False
            walker = filter_(in_genome, walker)

    try:
        filler = float(filler) if '.' in filler else int(filler)
    except ValueError:
        abort('Could not parse filler value: %s' % filler)

    write(fill(walker, regions=genome, filler=filler, only_edges=only_edges),
          name=name, description=description)


def derivative_track(track, method='forward', step=None, auto_step=False,
                     name=None, description=None):
    """
    Create derivative of a wiggle track.
    """
    if name is None and hasattr(track, 'name'):
        name = 'Derivative of %s' % track.name

    kwargs = {'step': step}
    if method == 'central':
        derivative = central_divided_difference
    elif method == 'backward':
        derivative = backward_divided_difference
        kwargs['auto_step'] = auto_step
    else:
        derivative = forward_divided_difference
        kwargs['auto_step'] = auto_step
    write(derivative(walk(track), **kwargs), name=name,
          description=description)


def plot_tracks(tracks, regions=None, genome=None, order_by='region',
                average_threshold=None, sharey=False, ylim=None, columns=None,
                pdf=None):
    """
    Visualize wiggle tracks in a plot.
    """
    def track_name(track, i):
        return track.name if hasattr(track, 'name') else 'track %i' % i

    # Prepend track name if we have more than one track.
    def rename(region, track, i):
        if len(tracks) > 1:
            return track_name(track, i) + ', ' + region
        return region

    if genome is not None:
        # Read genome from BED track and filter by specified regions. If we
        # have more than one track, have a region copy per track.
        genome = {rename(r, track, i): v
                  for r, v in read_regions(genome).items()
                  for i, track in enumerate(tracks)
                  if regions is None or r in regions}

    # Filter by specified regions.
    def filtered(walker):
        if regions is None:
            return walker
        return filter_(lambda (r, p, v): r in regions, walker)

    # Have all tracks concatenated in one walker.
    walker = ((rename(r, track, i), p, v)
              for i, track in enumerate(tracks)
              for r, p, v in filtered(walk(track)))

    fig, axes, rows, columns = plot(
        walker, regions=genome, order_by=order_by,
        average_threshold=average_threshold, sharey=sharey, ylim=ylim,
        columns=columns)

    if pdf:
        fig.set_size_inches(6 * columns, 3 * rows)
        fig.savefig(pdf, format='pdf')
    else:
        pyplot.show()


def coverage_track(track, threshold=None, name=None, description=None):
    """
    Create coverage BED track of a wiggle track.
    """
    if name is None and hasattr(track, 'name'):
        name = 'Coverage of %s' % track.name

    # Todo: Define coverage per region, like in `coverage-wiggle-to-bed` from
    #     bio-playground (https://github.com/martijnvermaat/bio-playground).
    walker = walk(track)
    if threshold is not None:
        walker = filter_(lambda (r, p, v): v >= threshold, walker)

    intervals.write(intervals.coverage(walker), name=name,
                    description=description)


def merge_tracks(tracks, merger='sum', custom_merger=None, no_indices=False,
                 name=None, description=None):
    """
    Merge any number of wiggle tracks in various ways.
    """
    if name is None and all(hasattr(track, 'name') for track in tracks):
        name = 'Merge of %s' % ', '.join(track.name for track in tracks)

    if custom_merger:
        # http://docs.python.org/2/reference/lexical_analysis.html#identifiers
        if re.match('[_a-zA-Z][_a-zA-Z0-9]*(\.[_a-zA-Z][_a-zA-Z0-9]*)+$',
                    custom_merger):
            # Importable definition, e.g. `package.module.merge_function`.
            module, name = custom_merger.rsplit('.', 1)
            merge_function = getattr(importlib.import_module(module), name)
        else:
            # Expression over `values`, e.g. `max(values)`.
            merge_function = eval('lambda values: ' + custom_merger)
    else:
        merge_function = mergers[merger]

    walkers = [walk(track, force_index=not no_indices)
               for track in tracks]
    write(merge(*walkers, merger=merge_function), name=name,
          description=description)


def distance_tracks(tracks, metric='a', threshold=None):
    """
    Calculate the distance between wiggle tracks.
    """
    # Todo: Cleanup this code.
    distances = distance(*tracks, metric=metrics[metric],
                         threshold=threshold)

    def name(index):
        return chr(ord('A') + index)

    try:
        sys.stdout.write(''.join('%s: %s\n' % (name(i), track.name)
                                 for i, track in enumerate(tracks)))
    except IOError:
        pass
    sys.stdout.write('\n   ')
    sys.stdout.write(' '.join('   %s ' % name(i)
                              for i in range(len(tracks))))
    sys.stdout.write('\n')
    sys.stdout.write(name(0) + '     x\n')
    for i in range(1, len(tracks)):
        sys.stdout.write('%s  ' % name(i))
        for j in range(0, i):
            sys.stdout.write(' %.3f' % distances[i, j])
        sys.stdout.write('   x\n')


def main():
    """
    Wiggelen command line interface.
    """
    parser = argparse.ArgumentParser(description=__doc__.split('\n\n')[0])
    subparsers = parser.add_subparsers(
        title='subcommands', dest='subcommand', help='subcommand help')

    p = subparsers.add_parser(
        'index', help='build index for wiggle track',
        description=index_track.__doc__.split('\n\n')[0])
    p.set_defaults(func=index_track)
    p.add_argument(
        'track', metavar='TRACK', type=argparse.FileType('r'),
        help='wiggle track')

    p = subparsers.add_parser(
        'sort', help='sort wiggle track regions alphabetically',
        description=sort_track.__doc__.split('\n\n')[0])
    p.set_defaults(func=sort_track)
    p.add_argument(
        'track', metavar='TRACK', type=argparse.FileType('r'),
        help='wiggle track')
    p.add_argument(
        '-n', '--name', dest='name', type=str,
        help='name to use for result track, displayed to the left of the '
        'track in the UCSC Genome Browser (default: Sorted TRACK)')
    p.add_argument(
        '-d', '--description', dest='description', type=str,
        help='description to use for result track, displayed as center label '
        'in the UCSC Genome Browser (default: no description)')

    p = subparsers.add_parser(
        'scale', help='scale values in a wiggle track',
        description=scale_track.__doc__.split('\n\n')[0])
    p.set_defaults(func=scale_track)
    p.add_argument(
        'track', metavar='TRACK', type=argparse.FileType('r'),
        help='wiggle track')
    p.add_argument(
        '-f', '--factor', dest='factor', type=float, default=0.1,
        help='scaling factor to use (default: %(default)s)')
    p.add_argument(
        '-n', '--name', dest='name', type=str,
        help='name to use for result track, displayed to the left of the '
        'track in the UCSC Genome Browser (default: Sorted TRACK)')
    p.add_argument(
        '-d', '--description', dest='description', type=str,
        help='description to use for result track, displayed as center label '
        'in the UCSC Genome Browser (default: no description)')

    p = subparsers.add_parser(
        'fill', help='fill undefined positions in a wiggle track',
        description=fill_track.__doc__.split('\n\n')[0],
        epilog='Note that the resulting track may be very large if '
        '--only-edges is not specified.')
    p.set_defaults(func=fill_track)
    p.add_argument(
        'track', metavar='TRACK', type=argparse.FileType('r'),
        help='wiggle track')
    p.add_argument(
        '-g', '--genome', dest='genome', type=argparse.FileType('r'),
        help='regions in BED format to fill (if not specified, first and '
        'last defined positions in the track define the regions)')
    p.add_argument(
        '-f', '--filler', dest='filler', default='0',
        help='value to use for undefined positions (default: %(default)s)')
    p.add_argument(
        '-e', '--only-edges', dest='only_edges', action='store_true',
        help='only fill the first and last of continuously undefined '
        'positions')
    p.add_argument(
        '-o', '--only-genome', dest='only_genome', action='store_true',
        help='only report positions in regions defined by the GENOME file')
    p.add_argument(
        '-n', '--name', dest='name', type=str,
        help='name to use for result track, displayed to the left of the '
        'track in the UCSC Genome Browser (default: Sorted TRACK)')
    p.add_argument(
        '-d', '--description', dest='description', type=str,
        help='description to use for result track, displayed as center label '
        'in the UCSC Genome Browser (default: no description)')

    p = subparsers.add_parser(
        'derivative', help='create derivative of a wiggle track',
        description=derivative_track.__doc__.split('\n\n')[0])
    p.set_defaults(func=derivative_track)
    p.add_argument(
        'track', metavar='TRACK', type=argparse.FileType('r'),
        help='wiggle track')
    p.add_argument(
        '-m', '--method', dest='method', type=str, default='forward',
        choices=('forward', 'backward', 'central'),
        help='type of divided difference method to use (default: %(default)s)')
    g = p.add_mutually_exclusive_group()
    g.add_argument(
        '-s', '--step', dest='step', type=int, default=None,
        help='restrict to positions that are this far apart (default: no '
        'restriction)')
    g.add_argument(
        '-a', '--auto-step', dest='auto_step', action='store_true',
        help='automatically set STEP to a value based on the first two '
        'positions in TRACK (default if METHOD is central and STEP is '
        'omitted)')
    p.add_argument(
        '-n', '--name', dest='name', type=str,
        help='name to use for result track, displayed to the left of the '
        'track in the UCSC Genome Browser (default: Sorted TRACK)')
    p.add_argument(
        '-d', '--description', dest='description', type=str,
        help='description to use for result track, displayed as center label '
        'in the UCSC Genome Browser (default: no description)')

    if plot is not None:
        p = subparsers.add_parser(
            'plot', description=plot_tracks.__doc__.split('\n\n')[0],
            help='visualize wiggle tracks in a plot (requires matplotlib)')
        p.set_defaults(func=plot_tracks)
        p.add_argument(
            'tracks', metavar='TRACK', type=argparse.FileType('r'), nargs='+',
            help='wiggle track')
        p.add_argument(
            '-r', '--regions', dest='regions', type=str, default=None,
            nargs='+', help='plot only these regions (default: plot all '
            'regions with data)')
        p.add_argument(
            '-g', '--genome', dest='genome', type=argparse.FileType('r'),
            help='regions with start and end positions in BED format')
        p.add_argument(
            '--order-by', dest='order_by', type=str, default='region',
            choices=['region', 'average', 'original'],
            help='when plotting multiple tracks and/or regions, order them '
            'by this key (default: region)')
        p.add_argument(
            '-t', '--avg-threshold', dest='average_threshold', type=float,
            default=None, help='Plot regions with average value above or '
            'equal to this threshold (default: plot regions that have data)')
        g = p.add_mutually_exclusive_group()
        g.add_argument(
            '-s', '--share-y', dest='sharey', action='store_true',
            help='when plotting multiple tracks and/or regions, share their '
            'y-axes')
        g.add_argument(
            '-y', '--y-lim', metavar=('MIN', 'MAX'), dest='ylim', nargs=2,
            type=float, help='set y-axis limits (default: automatically '
            'chosen)')
        p.add_argument(
            '-c', '--columns', dest='columns', type=int, default=None,
            help='when plotting multiple tracks and/or regions, use this '
            'many columns (default: automatically chosen)')
        p.add_argument(
            '-o', '--output', dest='pdf', type=argparse.FileType('wb'),
            default=None, help='output PDF file')

    p = subparsers.add_parser(
        'coverage', help='create coverage BED track of a wiggle track',
        description=coverage_track.__doc__.split('\n\n')[0])
    p.set_defaults(func=coverage_track)
    p.add_argument(
        'track', metavar='TRACK', type=argparse.FileType('r'),
        help='wiggle track')
    p.add_argument(
        '-t', '--threshold', dest='threshold', type=int, default=None,
        help='only include positions with this value or higher (default: no '
        'threshold)')
    p.add_argument(
        '-n', '--name', dest='name', type=str,
        help='name to use for result track, displayed to the left of the '
        'track in the UCSC Genome Browser (default: Sorted TRACK)')
    p.add_argument(
        '-d', '--description', dest='description', type=str,
        help='description to use for result track, displayed as center label '
        'in the UCSC Genome Browser (default: no description)')

    p = subparsers.add_parser(
        'merge', help='merge any number of wiggle tracks in various ways',
        description=merge_tracks.__doc__.split('\n\n')[0])
    p.set_defaults(func=merge_tracks)
    g = p.add_mutually_exclusive_group()
    g.add_argument(
        '-m', '--merger', dest='merger', choices=mergers, default='sum',
        help='merge operation (default: %(default)s)')
    g.add_argument(
        '-c', '--custom-merger', dest='custom_merger',
        help='custom Python merge operation, specified either by an '
        'expression over the list "values" (e.g., "max(values)"), or an '
        'importable name (e.g., "package.module.merger") that can be called '
        'with a list as argument')
    p.add_argument(
        '-x', '--no-indices', dest='no_indices', action='store_true',
        help='assume tracks are sorted, don\'t force building indices')
    p.add_argument(
        'tracks', metavar='TRACK', nargs='+', type=argparse.FileType('r'),
        help='wiggle track')
    p.add_argument(
        '-n', '--name', dest='name', type=str,
        help='name to use for result track, displayed to the left of the '
        'track in the UCSC Genome Browser (default: Sorted TRACK)')
    p.add_argument(
        '-d', '--description', dest='description', type=str,
        help='description to use for result track, displayed as center label '
        'in the UCSC Genome Browser (default: no description)')

    # Todo: Add additional information on the metrics (using the epilog
    # argument of the subparser).
    p = subparsers.add_parser(
        'distance', help='calculate the distance between wiggle tracks',
        description=distance_tracks.__doc__.split('\n\n')[0])
    p.set_defaults(func=distance_tracks)
    p.add_argument(
        '-m', dest='metric', choices=metrics, default='a',
        help='pairwise distance metric to use (default: %(default)s)')
    p.add_argument(
        '-t', dest='threshold', type=float, default=None,
        help='threshold for noise filter (default: no noise filter)')
    p.add_argument(
        'tracks', metavar='TRACK', nargs='+', type=argparse.FileType('r'),
        help='wiggle track')

    args = parser.parse_args()

    try:
        args.func(**{k: v for k, v in vars(args).items()
                     if k not in ('func', 'subcommand')})
    except IOError as e:
        abort(str(e))


if __name__ == '__main__':
    main()



"""
Plot wiggle tracks.

.. note:: This module depends on the :mod:`matplotlib` package.

.. moduleauthor:: Martijn Vermaat <martijn@vermaat.name>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import division

import collections
import itertools
import math

from matplotlib import pyplot

from .wiggle import fill


def plot(walker, regions=None, order_by='region', average_threshold=None,
         sharey=False, ylim=None, columns=None, line=True):
    """
    Visualize a wiggle track in a plot.

    For every region, a separate subplot is created.

    :arg walker: Generator yielding tuples of (region, position, value) per
        defined position.
    :type walker: generator(str, int, _)
    :arg regions: Dictionary with regions as keys and (start, stop) tuples as
        values. If not `None`, plot positions from start to stop (both
        including) in these regions. If `None`, plot positions in all
        regions between their first and last defined positions.
    :type regions: dict(str, (int, int))
    :arg order_by: Order the subplots by this key. Possible values are:
        - ``region``: Region name.
        - ``average``: Average value over all defined positions.
        - ``original``: Original position in `walker`.
    :type order_by: str
    :arg average_threshold: Include a subplot for every region with average
        value above or equal to this threshold. If `None`, only include those
        regions with data.
    :type average_threshold: float
    :arg sharey: Share y-axes in subplot arrangement.
    :type sharey: bool
    :arg ylim: Set the y-limits of all subplots.
    :type ylim: tuple(float)
    :arg columns: Number of columns to use in subplot arrangement. If `None`,
        a suitable number of columns is chosen automatically.
    :type columns: int
    :arg line: Connect values to create a lineplot. Undefined positions are
        assumed to be `0`.
    :type line: bool

    :return: A tuple containing a matplotlib figure object, a list of
        subplots, the number of rows and the number of columns.
    :rtype: matplotlib.figure.Figure, list(matplotlib.axes.AxesSubplot), int,
        int

    .. note:: This function loads the walker in memory and as such is not
        appropriate for use on very large datasets.
    """
    fig = pyplot.figure(tight_layout=True)
    subplots = collections.OrderedDict()

    if line:
        walker = fill(walker, regions=regions, filler=0, only_edges=True)

    # Keep track of the last added subplot for sharing the Y axis.
    ax = None

    for region, w in itertools.groupby(walker, lambda (r, p, v): r):
        if regions is not None:
            try:
                start, stop = regions[region]
                positions, values = zip(*[(p, v) for r, p, v in w
                                          if start < p <= stop])
            except KeyError:
                continue
        else:
            positions, values = zip(*[(p, v) for r, p, v in w])
            start, stop = min(positions), max(positions)

        average = sum(v for v in values if v is not None) / (stop - start + 1)
        if average_threshold is not None and average < average_threshold:
            continue

        # Temporarily add the subplot at location 111, we'll change that
        # once we know how many subplots there are.
        ax = fig.add_subplot(111, label=region, sharey=ax if sharey else None)
        if line:
            ax.plot(positions, values, color='b')
        else:
            ax.plot(positions, values, color='b', linestyle='None', marker=',')
        ax.plot([start, stop], [average, average], color='r', linestyle='--')
        ax.set_xlim(start, stop)
        if ylim:
            ax.set_ylim(ylim)
        ax.set_title(region)
        ax.legend([], title='avg=%.2f' % average, loc='best')
        subplots[region] = average, ax

    # If the threshold is zero, also include every region from the defined
    # region that has no data.
    if average_threshold == 0:
        for region in regions:
            if region not in subplots:
                ax = fig.add_subplot(111, label=region,
                                     sharey=ax if sharey else None)
                ax.set_xlim(*regions[region])
                ax.set_title(region)
                ax.text(0.5, 0.5, 'no data', ha='center', va='center',
                        transform=ax.transAxes, size=24, alpha=0.5)
                subplots[region] = 0, ax

    # Get all the axes in the specified order.
    if order_by == 'region':
        # Alphabetically by region name.
        axes = [subplots[r][1] for r in sorted(subplots)]
    elif order_by == 'average':
        # By average value.
        axes = [subplots[r][1] for r in sorted(subplots, key=subplots.get,
                                               reverse=True)]
    else:
        # By original order in de walker.
        axes = [s[1] for s in subplots.values()]

    columns = columns or int(math.sqrt(len(axes)))
    rows = int((len(axes) - 1) / columns + 1)

    for i, ax in enumerate(axes):
        ax.change_geometry(rows, columns, i + 1)

    return fig, axes, rows, columns



"""
Helper functions for parsing wiggle tracks.

.. moduleauthor:: Martijn Vermaat <martijn@vermaat.name>

.. Licensed under the MIT license, see the LICENSE file.
"""


from collections import namedtuple


class ParseError(Exception):
    """
    Raised if a wiggle track cannot be parsed.
    """
    pass


# Line types.
class LineType(object):
    NONE, REGION, DATA = range(3)


# Track modes.
class Mode(object):
    VARIABLE, FIXED = range(2)


Data = namedtuple('Data', 'position span value')


# Create a new object encapsulating state for the parse function.
def create_state():
    return dict(mode=Mode.VARIABLE, span=1, start=None, step=None)


def parse(line, state):
    # Parse a line and return a tuple (line_type, data). The state dictionary
    # is modified and should be passed as such with the next call.

    # As an optimization, we first check for the common case of a line with
    # data. It must always start with a number (either position or value).
    if line[0] in '0123456789.':
        if state['mode'] == Mode.VARIABLE:
            try:
                position, value = line.split()
                return LineType.DATA, Data(
                    position=int(position),
                    span=state['span'],
                    value=float(value) if '.' in value else int(value))
            except ValueError:
                raise ParseError('Could not parse line: %s' % line)

        if state['mode'] == Mode.FIXED:
            try:
                position = state['start']
                state['start'] += state['step']
                return LineType.DATA, Data(
                    position=position,
                    span=state['span'],
                    value=float(line) if '.' in line else int(line))
            except ValueError:
                raise ParseError('Could not parse line: %s' % line)

    if (line[:7] == 'browser' or line[:5] == 'track' or line[0] == '#' or
        line in ('\n', '\r\n', '\r')):
        # As far as I can see empty lines and comments are not allowed
        # by the spec, but I guess they exist in the real world.
        return LineType.NONE, None

    if line[:12] == 'variableStep':
        try:
            fields = dict(map(lambda field: field.split('='),
                              line[len('variableStep'):].split()))
            state['mode'] = Mode.VARIABLE
            state['span'] = int(fields.get('span', 1))
            return LineType.REGION, fields['chrom']
        except (ValueError, KeyError):
            raise ParseError('Could not parse line: %s' % line)

    if line[:9] == 'fixedStep':
        try:
            fields = dict(map(lambda field: field.split('='),
                              line[len('fixedStep'):].split()))
            state['mode'] = Mode.FIXED
            state['start'] = int(fields['start'])
            state['span'] = int(fields.get('span', 1))
            # Though not valid by the specification, we accept fixedStep
            # definitions without a step argument (it's also accepted by the
            # UCSC Genome Browser).
            # Issue: https://github.com/martijnvermaat/wiggelen/issues/1
            state['step'] = int(fields.get('step', min(1, state['span'])))
            return LineType.REGION, fields['chrom']
        except (ValueError, KeyError):
            raise ParseError('Could not parse line: %s' % line)

    raise ParseError('Could not parse line: %s' % line)



"""
Calculate the distance between two wiggle tracks using a metric designed for
multisets.

This module can be used to assess similarity of next generation sequencing
datasets. A multiset distance measure is used for pairwise comparison of
genomic information as provided by wiggle tracks.

The algorithm can be parameterized by a pairwise distance metric. Four of
these metrics are predefined in :attr:`metrics`:

Metric ``a``: :math:`\\frac{|x - y|}{(x + 1) (y + 1)}`

Metric ``b``: :math:`\\frac{|x - y|}{x + y + 1}`

Metric ``c``: :math:`\\frac{\\text{max}(x, y) \, |x - y|}{(x^2 + 1) (y^2 + 1)}`

Metric ``d``: :math:`\\frac{|x - y|}{\\text{max}(x, y) + 1}`

.. note:: These metrics are ill-defined on the interval (0, 1) so we scale all
       values if necessary.

.. moduleauthor:: Martijn Vermaat <martijn@vermaat.name>
.. moduleauthor:: Jeroen Laros <j.f.j.laros@lumc.nl>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import division

from collections import defaultdict
import itertools
import sys

from .wiggle import walk
from .index import Field, index
from .merge import merge


# Todo: Give these metrics a name.
_metric_a = lambda x, y: abs(x - y) / ((x + 1) * (y + 1))
_metric_b = lambda x, y: abs(x - y) / (x + y + 1)
_metric_c = lambda x, y: ((max(x, y) * abs(x - y))
                          / (((x * x) + 1) * ((y * y) + 1)))
_metric_d = lambda x, y: abs(x - y) / (max(x, y) + 1)


#: Predefined pairwise distance metrics. See :mod:`wiggelen.distance` for
#: their definition.
metrics = {'a': _metric_a,
           'b': _metric_b,
           'c': _metric_c,
           'd': _metric_d}


def normalize(*values):
    """
    Normalize values relative to the smallest value.

    :arg values: List of values.
    :type values: list(float)

    :return: Scale the values such that the minimum is 1.
    :rtype: list(float)
    """
    min_value = min(*values)
    if not min_value:
        return [1 for _ in values]
    return [value / min_value for value in values]


def matrix(size, reflexive=False, symmetric=False):
    """
    Create all coordinates in a square matrix.

    With the default `False` value for `reflexive` and `symmetric`, include
    only the coordinates below the diagonal.

    :arg size: Width and height of the matrix.
    :type size: int
    :arg reflexive: Include coordinates on (`x`, `x`) diagonal.
    :type reflexive: bool
    :arg symmetric: Include coordinates (`x`, `y`) above the diagonal (where
        `x < y`).
    :type symmetric: bool

    :return: All coordinates in the matrix as tuples.
    :rtype: list(int, int)

    Examples::

        >>> matrix(5)
        [(1, 0),
         (2, 0), (2, 1),
         (3, 0), (3, 1), (3, 2),
         (4, 0), (4, 1), (4, 2), (4, 3)]
        >>> matrix(5, reflexive=True)
        [(0, 0),
         (1, 0), (1, 1),
         (2, 0), (2, 1), (2, 2),
         (3, 0), (3, 1), (3, 2), (3, 3),
         (4, 0), (4, 1), (4, 2), (4, 3), (4, 4)]
        >>> matrix(5, symmetric=True)
        [        (0, 1), (0, 2), (0, 3), (0, 4),
         (1, 0),         (1, 2), (1, 3), (1, 4),
         (2, 0), (2, 1),         (2, 3), (2, 4),
         (3, 0), (3, 1), (3, 2),         (3, 4),
         (4, 0), (4, 1), (4, 2), (4, 3)        ]
    """
    return [(i, j) for i in range(0, size) for j in
            itertools.chain(range(i + 1 if reflexive else i),
                            range(i + 1, size if symmetric else 0))]


def distance(*tracks, **options):
    """
    Calculate the pairwise distances between wiggle tracks.

    :arg tracks: List of wiggle tracks.
    :type walkers: list(file)
    :arg metric: Pairwise distance metric (default: a).
    :type merger: function(float, float -> float)
    :arg threshold: Threshold for noise filter (default: no noise filter)
    :type threshold: float

    :return: Pairwise distances between `tracks` as a mapping from
        coordinates in the distance matrix to their values.
    :rtype: dict((int, int), float)

    .. todo:: Check where this goes wrong if we cannot .seek() the tracks.
    .. todo:: Calculate weights per region instead of over the entire track.
    """
    metric = options.get('metric', metrics['a'])
    threshold = options.get('threshold')

    # We construct a list of comparisons for the merger, where each comparison
    # is a tuple of (left, right, weight_left, weight_right).
    comparisons = []

    if threshold:
        field_suffix = '-threshold-%s' % str(threshold)
        noise_filter = lambda value: max(value - threshold, 0)

        def sum_func(acc, value, span):
            return acc + noise_filter(value) * span

        def min_func(acc, value, span):
            filtered = noise_filter(value)
            if filtered > 0:
                return min(acc, noise_filter(value))
            return acc

        fields = [Field('sum' + field_suffix, float, 0, sum_func),
                  Field('posmin' + field_suffix, float, sys.float_info.max,
                        min_func)]

    else:
        field_suffix = ''
        noise_filter = lambda value: value
        fields = []

    summaries = [index(track, force=True, fields=fields)[0]['_all']
                 for track in tracks]

    # Our metrics are undifined on the (0, 1) interval, so if the positive
    # minimum over all tracks is < 1 we upscale everything.
    min_value = min(summary['posmin' + field_suffix] for summary in summaries)
    scale = 1 / min_value if 0 < min_value < 1 else 1

    # Based on the sums of all values in each track we define weights.
    sums = [summary['sum' + field_suffix] for summary in summaries]
    for left, right in matrix(len(tracks)):
        weight_right, weight_left = normalize(sums[left], sums[right])
        comparisons.append( (left, right, weight_left, weight_right) )

    # The merger returns a matrix of values, one for each pairwise comparison.
    def merger(values):
        results = {}
        for left, right, weight_left, weight_right in comparisons:
            value_left, value_right = values[left], values[right]
            if value_left is None and value_right is None:
                result = None
            else:
                x = (weight_left * scale * noise_filter(value_left)
                     if value_left else 0)
                y = (weight_right * scale * noise_filter(value_right)
                     if value_right else 0)
                result = metric(x, y) if x or y else None
            results[left, right] = result
        return results

    # Indexed walkers.
    walkers = [walk(track, force_index=True) for track in tracks]

    # Aggregate results.
    totals = defaultdict(lambda: 0)
    counts = defaultdict(lambda: 1)
    for _, _, values in merge(*walkers, merger=merger):
        for comparison, value in values.items():
            if value is not None:
                totals[comparison] += value
                counts[comparison] += 1

    # Create the distance matrix by taking the averages.
    distances = {}
    for comparison in counts:
        distances[comparison] = totals[comparison] / counts[comparison]
    return distances



"""
Merge any number of wiggle tracks in various ways.

The algorithm can be parameterized by a merge operation. Four of these
operations are predefined in :attr:`mergers`:

Merger ``sum``: Compute the sum of all values.

Merger ``mean``: Compute the mean of all values (and use 0 for undefined
values).

Merger ``count``: Compute the number of defined values.

Merger ``minus``: Subtract the second value from the first (and use 0 for
undefined values). Only defined on exactly two values.

Merger ``min``: Compute the minimum of all values (and use 0 for undefined
values).

Merger ``max``: Compute the maximum of all values (and use 0 for undefined
values).

Merger ``div``: Divide the second value by the first (and use 0 for
undefined values). Only defined on exactly two values.

Merger ``intersect``: Return the first value is the second value is defined
and non-zero (and use 0 for undefined values). Only defined on exactly two
values.

Merger ``ctz``: Select the value closest to 0. (and use 0 if there is a mix of
positive and negative values).

.. moduleauthor:: Martijn Vermaat <martijn@vermaat.name>
.. moduleauthor:: Jeroen F.J. Laros <J.F.J.Laros@lumc.nl>

.. Licensed under the MIT license, see the LICENSE file.
"""


from __future__ import division

from .wiggle import zip_


# Compute the sum of all values.
_merger_sum = lambda vs: sum(v for v in vs if v is not None)

# Compute the mean of all values (and use 0 for undefined values).
# Todo: Also provide a mean over only the defined values.
_merger_mean = lambda vs: _merger_sum(vs) / len(vs)

# Compute the number of defined values.
_merger_count = lambda vs: sum(1 for v in vs if v is not None)

# Substract the second from the first (and use 0 for undefined values).
_merger_minus = lambda vs: (vs[0] or 0) - (vs[1] or 0)

# Compute the minimum of all values (and use 0 for undefined values).
_merger_min = lambda vs: min((v or 0) for v in vs)

# Compute the minimum of all values (and use 0 for undefined values).
_merger_max = lambda vs: max((v or 0) for v in vs)

# Divide the second by the first (and use 0 for undefined values).
_merger_div = lambda vs: (vs[0] or 0) / (vs[1] or 1)

# Return the first if the second is defined and non-zero (and use 0 for
# undefined values).
_merger_intersect = lambda vs: (vs[0] or 0) if vs[1] else 0

def _merger_ctz(vs):
    """
    Select the value closest to 0. (and use 0 if there is a mix of positive and
    negative values).

    :arg vs: List of integer or None values.
    :type vs: list
    """
    minimum = min(vs)
    if minimum >= 0:
        return minimum

    maximum = max(vs)
    if maximum <= 0:
        return maximum

    return 0


#: Predefined mergers. See :mod:`wiggelen.merge` for their definition.
mergers = {'sum':       _merger_sum,
           'mean':      _merger_mean,
           'count':     _merger_count,
           'minus':     _merger_minus,
           'min':       _merger_min,
           'max':       _merger_max,
           'div':       _merger_div,
           'intersect': _merger_intersect,
           'ctz':       _merger_ctz}


def merge(*walkers, **options):
    """
    Merge wiggle tracks.

    This assumes the walkers have their regions in the same order. You can
    force this by using indices. Example::

        >>> from wiggelen import walk
        >>> walkers = [walk(open(track), force_index=True)
        ...            for track in ('a.wig', 'b.wig', 'c.wig')]
        >>> for x in merge(*walkers):
        ...     x
        ...
        ('18', 8, 849.0)
        ('18', 9, 987.0)
        ('MT', 1, 820.0)

    :arg walkers: List of generators yielding tuples of (region, position,
        value) per defined position.
    :type walkers: list(generator(str, int, _))
    :keyword merger: Merge operation (default: sum).
    :type merger: function(list(_) -> _)

    :return: Tuples of (region, position, merged value) per defined position
        in `walkers`.
    :rtype: generator(str, int, _)
    """
    # Todo: Would it be better to also pass region/position to the merger?
    merger = options.get('merger', mergers['sum'])

    for region, position, values in zip_(*walkers):
        yield region, position, merger(values)



"""
Get covered intervals from wiggle tracks and write to BED format.

.. moduleauthor:: Martijn Vermaat <martijn@vermaat.name>

.. Licensed under the MIT license, see the LICENSE file.
"""


import sys


def coverage(walker):
    """
    Get intervals of consecutively defined positions from a walker.

    :arg walker: Tuple of `(region, position, value)` per defined position.
    :type walker: generator(str, int, _)

    :return: Tuples of `(region, begin, end)` per position where `begin` and
        `end` are one-based and inclusive.
    :rtype: generator(str, int, int)

    Example::

        >>> for x in coverage(walk(open('a.wig'))):
        ...     x
        ...
        ('MT', 3, 3)
        ('MT', 5, 20)
        ('MT', 400, 420)
    """
    interval = None

    for region, position, _ in walker:
        if interval is not None:
            if region != interval[0] or position != interval[2] + 1:
                yield interval
                interval = None

        if interval is None:
            interval = region, position, position
        else:
            interval = interval[0], interval[1], position

    # Backlog.
    if interval is not None:
        yield interval


def write(intervals, track=sys.stdout, name=None, description=None):
    """
    Write intervals to a bed track.

    :arg intervals: Tuples of (region, begin, end) per interval.
    :type intervals: generator(str, int, int)
    :arg track: Writable file handle.
    :type track: file
    :arg name: Optional track name (displayed to the left of the track in the
        UCSC Genome Browser).
    :type name: str
    :arg description: Optional track description (displayed as center label in
        the UCSC Genome Browser).
    :type description: str

    Example::

       >>> write(coverage(walk(open('a.wig'))), name='My example')
       track name="My example"
       MT 2 3
       MT 4 20
       MT 399 420
    """
    header = 'track'
    if name is not None:
        header += ' name="%s"' % name
    if description is not None:
        header += ' description="%s"' % description
    header += '\n'
    track.write(header)

    for interval in intervals:
        track.write('%s\t%i\t%i\n' % interval)
